@@ -1,7 +1,7 @@
 /**************************************************************************
  *   nano.h  --  This file is part of GNU nano.                           *
  *                                                                        *
- *   Copyright (C) 1999-2011, 2013-2019 Free Software Foundation, Inc.    *
+ *   Copyright (C) 1999-2011, 2013-2020 Free Software Foundation, Inc.    *
  *   Copyright (C) 2014-2017 Benno Schulenberg                            *
  *                                                                        *
  *   GNU nano is free software: you can redistribute it and/or modify     *
@@ -19,9 +19,6 @@
  *                                                                        *
  **************************************************************************/
 
-#ifndef NANO_H
-#define NANO_H 1
-
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
@@ -181,10 +178,12 @@ typedef enum {
 /* Structure types. */
 #ifdef ENABLE_COLOR
 typedef struct colortype {
+	short id;
+		/* An ordinal number (if this color combo is for a multiline regex). */
 	short fg;
-		/* This syntax's foreground color. */
+		/* This combo's foreground color. */
 	short bg;
-		/* This syntax's background color. */
+		/* This combo's background color. */
 	short pairnum;
 		/* The pair number for this foreground/background color combination. */
 	int attributes;
@@ -194,9 +193,7 @@ typedef struct colortype {
 	regex_t *end;
 		/* The compiled regular expression for 'end=', if any. */
 	struct colortype *next;
-		/* Next set of colors. */
-	int id;
-		/* Basic id for assigning to lines later. */
+		/* Next color combination. */
 } colortype;
 
 typedef struct regexlisttype {
@@ -234,6 +231,8 @@ typedef struct syntaxtype {
 		/* The list of libmagic results that this syntax applies to. */
 	char *linter;
 		/* The command with which to lint this type of file. */
+	char *formatter;
+		/* The command with which to format/modify/arrange this type of file. */
 	char *tab;
 		/* What the Tab key should produce; NULL for default behavior. */
 #ifdef ENABLE_COMMENT
@@ -242,7 +241,7 @@ typedef struct syntaxtype {
 #endif
 	colortype *color;
 		/* The colors and their regexes used in this syntax. */
-	int nmultis;
+	short nmultis;
 		/* How many multiline regex strings this syntax has. */
 	struct syntaxtype *next;
 		/* Next syntax. */
@@ -311,9 +310,9 @@ typedef struct undostruct {
 		/* The operation type that this undo item is for. */
 	int xflags;
 		/* Some flag data to mark certain corner cases. */
-	ssize_t lineno;
+	ssize_t head_lineno;
 		/* The line number where the operation began or ended. */
-	size_t begin;
+	size_t head_x;
 		/* The x position where the operation began or ended. */
 	char *strdata;
 		/* String data to help restore the affected line. */
@@ -325,9 +324,9 @@ typedef struct undostruct {
 		/* Undo info specific to groups of lines. */
 	linestruct *cutbuffer;
 		/* A copy of the cutbuffer. */
-	ssize_t mark_begin_lineno;
+	ssize_t tail_lineno;
 		/* Mostly the line number of the current line; sometimes something else. */
-	size_t mark_begin_x;
+	size_t tail_x;
 		/* The x position corresponding to the above line number. */
 	struct undostruct *next;
 		/* A pointer to the undo item of the preceding action. */
@@ -423,8 +422,6 @@ typedef struct rcoption {
 typedef struct keystruct {
 	const char *keystr;
 		/* The string that describes the keystroke, like "^C" or "M-R". */
-	bool meta;
-		/* Whether this is a Meta keystroke. */
 	int keycode;
 		/* The integer that, together with meta, identifies the keystroke. */
 	int menus;
@@ -496,7 +493,7 @@ enum
 	CASE_SENSITIVE,
 	CONSTANT_SHOW,
 	NO_HELP,
-	SUSPEND,
+	SUSPENDABLE,
 	NO_WRAP,
 	AUTOINDENT,
 	VIEW_MODE,
@@ -568,8 +565,6 @@ enum
 #endif
 
 /* Basic control codes. */
-#define BS_CODE   0x08
-#define TAB_CODE  0x09
 #define ESC_CODE  0x1B
 #define DEL_CODE  0x7F
 
@@ -611,6 +606,9 @@ enum
 /* A special keycode for when <Tab> is pressed while the mark is on. */
 #define INDENT_KEY 0x4F1
 
+/* A special keycode to signal the beginning and end of a bracketed paste. */
+#define BRACKETED_PASTE_MARKER 0x4FB
+
 #ifdef USE_SLANG
 #ifdef ENABLE_UTF8
 #define KEY_BAD 0xFF  /* Clipped error code. */
@@ -625,11 +623,11 @@ enum
 #define KEY_WINCH -2
 
 /* Some extra flags for the undo function. */
-#define WAS_FINAL_BACKSPACE   (1<<1)
+#define WAS_BACKSPACE_AT_EOF  (1<<1)
 #define WAS_WHOLE_LINE        (1<<2)
-#define WAS_FINAL_LINE        (1<<3)
+#define INCLUDED_LAST_LINE    (1<<3)
 #define MARK_WAS_SET          (1<<4)
-#define WAS_MARKED_FORWARD    (1<<5)
+#define WAS_MARKED_BACKWARDS  (1<<5)
 #endif /* !NANO_TINY */
 
 /* The default number of columns from end of line where wrapping occurs. */
@@ -650,5 +648,3 @@ enum
 
 /* The largest size_t number that doesn't have the high bit set. */
 #define HIGHEST_POSITIVE ((~(size_t)0) >> 1)
-
-#endif /* !NANO_H */
